# A tool to upload and manage archives in AWS Glacier Vaults.
# Copyright (C) 2023 Trapsilo P. Bumi tbumi@thpd.io
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import concurrent.futures
import math
import os.path
import tarfile
import tempfile
import threading
import traceback

import boto3
import click

from .utils.tree_hash import calculate_total_tree_hash, calculate_tree_hash

SINGLE_UPLOAD_THRESHOLD_BYTES = 100 * 1024 * 1024  # 100 MB
MAX_UPLOAD_ATTEMPTS = 10  # 10 retries for each part before failing

# ref: https://docs.aws.amazon.com/amazonglacier/latest/dev/uploading-archive-mpu.html
MAX_NUMBER_OF_PARTS = 10000
MIN_PART_SIZE_MB = 1
MAX_PART_SIZE_MB = 4 * 1024


def upload_archive(
    vault_name, file_name, arc_desc, part_size_mb, num_threads, upload_id
):
    glacier = boto3.client("glacier")

    if part_size_mb < MIN_PART_SIZE_MB or part_size_mb > MAX_PART_SIZE_MB:
        raise click.ClickException(
            f"part-size must be between {MIN_PART_SIZE_MB} and {MAX_PART_SIZE_MB} MB"
        )
    if not math.log2(part_size_mb).is_integer():
        raise click.ClickException("part-size must be a power of 2")

    file_to_upload = None
    try:
        if len(file_name) > 1 or os.path.isdir(file_name[0]):
            click.echo("Consolidating files into a .tar archive...")
            file_to_upload = tempfile.TemporaryFile()
            with tarfile.open(fileobj=file_to_upload, mode="w:xz") as tar:
                for filename in file_name:
                    tar.add(filename)
            click.echo("Files consolidated.")
        else:
            file_to_upload = open(file_name[0], mode="rb")
            click.echo("Opened single file.")

        file_size_bytes = file_to_upload.seek(0, 2)
        file_to_upload.seek(0, 0)  # return file pointer to start of file

        if file_size_bytes < SINGLE_UPLOAD_THRESHOLD_BYTES:
            click.echo(
                f"File size is less than {SINGLE_UPLOAD_THRESHOLD_BYTES:,} bytes. "
                "Uploading in one request..."
            )
            response = glacier.upload_archive(
                vaultName=vault_name, archiveDescription=arc_desc, body=file_to_upload
            )

            click.echo("Uploaded.")
            click.echo(f"Glacier tree hash: {response['checksum']}")
            click.echo(f"Location: {response['location']}")
            click.echo(f"Archive ID: {response['archiveId']}")
        else:
            if (
                math.ceil(file_size_bytes / (part_size_mb * 1024 * 1024))
                > MAX_NUMBER_OF_PARTS
            ):
                target_part_size = file_size_bytes / (10000 * 1024 * 1024)
                new_part_size = MIN_PART_SIZE_MB
                while new_part_size < target_part_size:
                    # find the nearest power of 2 larger than the target part size
                    new_part_size *= 2
                    if new_part_size > MAX_PART_SIZE_MB:
                        raise click.ClickException(
                            "Archive/upload size too large (more than 40 TB)"
                        )
                click.confirm(
                    "Maximum number of parts exceeded, would you like to "
                    f"switch to {new_part_size} MB part size?",
                    default=True,
                    abort=True,
                )
                part_size_mb = new_part_size

            multipart_upload(
                glacier,
                upload_id,
                vault_name,
                arc_desc,
                file_size_bytes,
                part_size_mb * 1024 * 1024,
                file_to_upload,
                num_threads,
            )
    finally:
        if file_to_upload is not None:
            file_to_upload.close()

    click.echo("Done.")


def multipart_upload(
    glacier,
    upload_id,
    vault_name,
    arc_desc,
    file_size_bytes,
    part_size_bytes,
    file_to_upload,
    num_threads,
):
    part_list = {}  # map of byte_start -> checksum
    for byte_start in range(0, file_size_bytes, part_size_bytes):
        part_list[byte_start] = None
    num_parts = len(part_list)

    if upload_id is None:
        click.echo("Initiating multipart upload...")
        response = glacier.initiate_multipart_upload(
            vaultName=vault_name,
            archiveDescription=arc_desc,
            partSize=str(part_size_bytes),
        )
        upload_id = response["uploadId"]

        click.echo(
            f"File size is {file_size_bytes:,} bytes. "
            f"Will upload in {num_parts:,} parts."
        )
    else:
        click.echo(f"Resuming upload with id {upload_id}...")

        click.echo("Fetching already uploaded parts...")
        try:
            paginator = glacier.get_paginator("list_parts")
            response = paginator.paginate(vaultName=vault_name, uploadId=upload_id)
            parts = list(response.search("Parts"))
        except glacier.exceptions.ResourceNotFoundException as e:
            raise click.ClickException(e.response["Error"]["Message"])

        with click.progressbar(parts, label="Verifying uploaded parts") as bar:
            for part_data in bar:
                byte_start = int(part_data["RangeInBytes"].partition("-")[0])
                file_to_upload.seek(byte_start)
                part = file_to_upload.read(part_size_bytes)
                checksum = calculate_tree_hash(part, part_size_bytes)

                if checksum == part_data["SHA256TreeHash"]:
                    part_list[byte_start] = checksum

    click.echo("Spawning threads...")
    with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor:
        fileblock = threading.Lock()
        futures_list = {}
        for byte_pos in [
            part for part, checksum in part_list.items() if checksum is None
        ]:
            future = executor.submit(
                upload_part,
                byte_pos,
                vault_name,
                upload_id,
                part_size_bytes,
                file_to_upload,
                file_size_bytes,
                num_parts,
                glacier,
                fileblock,
            )
            futures_list[future] = byte_pos
        done, not_done = concurrent.futures.wait(
            futures_list, return_when=concurrent.futures.FIRST_EXCEPTION
        )
        if len(not_done) > 0:
            # an exception occurred
            for future in not_done:
                future.cancel()
            for future in done:
                exc = future.exception()
                if exc is not None:
                    exc_string = "".join(traceback.format_exception(exc))
                    click.secho(f"Exception occurred: {exc_string}", err=True, fg="red")
            click.echo(f"Upload can still be resumed. Upload ID: {upload_id}")
            raise click.Abort
        else:
            # all threads completed without raising an Exception
            for future in done:
                byte_start = futures_list[future]
                part_list[byte_start] = future.result()

    total_tree_hash = calculate_total_tree_hash(list(part_list.values()))

    click.echo("Completing multipart upload...")
    response = glacier.complete_multipart_upload(
        vaultName=vault_name,
        uploadId=upload_id,
        archiveSize=str(file_size_bytes),
        checksum=total_tree_hash,
    )
    click.echo("Upload successful.")
    click.echo(f"Calculated total tree hash: {total_tree_hash}")
    click.echo(f"Glacier total tree hash: {response['checksum']}")
    click.echo(f"Location: {response['location']}")
    click.echo(f"Archive ID: {response['archiveId']}")


def upload_part(
    start_pos,
    vault_name,
    upload_id,
    part_size_bytes,
    fp,
    file_size_bytes,
    num_parts,
    glacier,
    fileblock,
):
    with fileblock:
        fp.seek(start_pos)
        part = fp.read(part_size_bytes)

    end_pos = start_pos + len(part) - 1
    range_header = f"bytes {start_pos}-{end_pos}/{file_size_bytes}"
    part_num = start_pos // part_size_bytes
    percentage = part_num / num_parts
    checksum = calculate_tree_hash(part, part_size_bytes)

    click.echo(f"Uploading part {part_num + 1} of {num_parts}... ({percentage:.2%})")

    error = None
    for _ in range(MAX_UPLOAD_ATTEMPTS):
        try:
            response = glacier.upload_multipart_part(
                vaultName=vault_name, uploadId=upload_id, range=range_header, body=part
            )
            if checksum != response["checksum"]:
                raise Exception("Local checksum does not match Glacier checksum")

            # upload success, exit loop
            break
        except Exception as e:
            click.secho(f"Upload error: {e}", err=True, fg="red")
            click.echo(f"Trying again. Part {part_num + 1}")
            error = e
    else:
        click.secho(
            f"After {MAX_UPLOAD_ATTEMPTS} attempts, "
            f"still failed to upload part. Aborting upload of part {part_num + 1}.",
            err=True,
            fg="red",
        )
        if error is not None:
            raise error
        else:
            raise RuntimeError()

    del part  # freeing memory
    return checksum



# A tool to upload and manage archives in AWS Glacier Vaults.
# Copyright (C) 2023 Trapsilo P. Bumi tbumi@thpd.io
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import click

from . import archives, inventories, upload


@click.group()
def glacier_cli():
    pass


@glacier_cli.command(name="upload")
@click.option(
    "-d",
    "--arc-desc",
    default="",
    help="The archive description to help identify archives later",
)
@click.option(
    "-p",
    "--part-size",
    "part_size_mb",
    type=int,
    default=8,
    help=(
        "The part size for multipart upload, in megabytes as a power of 2"
        "(e.g. 1, 2, 4, 8). Default: 8"
    ),
)
@click.option(
    "-t",
    "--num-threads",
    type=int,
    default=((os.cpu_count() or 1) * 5),
    help=(
        "The amount of worker threads concurrently uploading data. "
        "Default: Number of CPUs multiplied by 5"
    ),
)
@click.option("-u", "--upload-id", help="If provided, will resume upload with this ID.")
@click.argument("vault_name")
@click.argument("file_name", nargs=-1, type=click.Path(exists=True))
def upload_archive(**args):
    """
    Uploads FILE_NAME to an AWS Glacier Vault named VAULT_NAME. If FILE_NAME is more
    than 1, they will be consolidated into a tar file before uploaded. If FILE_NAME
    is a directory, the directory and its contents will be consolidated into a single
    tar file and uploaded.
    """
    return upload.upload_archive(**args)


@glacier_cli.group(name="inventory")
def inventory_group():
    """
    Initiate or get an inventory retrieval job.
    """
    pass


@inventory_group.command(name="init-retrieval")
@click.option(
    "-f",
    "--format",
    default="JSON",
    type=click.Choice(["CSV", "JSON"]),
    help="Format of the inventory to request from glacier. Default: JSON",
)
@click.option("-d", "--description", help="Description of this job")
@click.argument("vault_name")
def init_inventory_retrieval(vault_name, format, description):
    """
    Initiate inventory retrieval for all archives in VAULT_NAME.

    This function tells AWS to start creating a list of all archives that exist
    in the vault. Because of the way archives are stored in glacier, this list
    cannot be available immediately. The inventory will be completed between
    several minutes to several hours. To check and retrieve the status of
    this job, run get-inventory with the vault name and job ID
    returned by this function.
    """
    return inventories.init_retrieval(vault_name, format, description)


@inventory_group.command(name="get")
@click.argument("vault_name")
@click.argument("job_id")
def get_inventory(vault_name, job_id):
    """
    Get the output of an inventory retrieval job identified by JOB_ID in VAULT_NAME.

    The inventory job must have already been initialized by init-inventory-retrieval.
    """
    return inventories.get(vault_name, job_id)


@glacier_cli.group(name="archive")
def archive_group():
    """
    Initiate or get an archive retrieval job.
    """
    pass


@archive_group.command(name="init-retrieval")
@click.option("-d", "--description", help="Description of this job")
@click.argument("vault_name")
@click.argument("archive_id")
def init_archive_retrieval(vault_name, archive_id, description):
    """
    Initiate retrieval for the archive designated by ARCHIVE_ID in VAULT_NAME.

    To retrieve an archive from glacier, use this function. Because of the
    way archives are stored in glacier, it will take some time to retrieve
    the archive. To check the status and retrieve the archive, run
    glacier-get-archive with the vault name and job ID returned by this function.
    """
    return archives.init_retrieval(vault_name, archive_id, description)


@archive_group.command(name="get")
@click.argument("vault_name")
@click.argument("job_id")
@click.argument(
    "file_name",
    type=click.Path(dir_okay=False, writable=True),
)
def get_archive(vault_name, job_id, file_name):
    """
    Get the output of an archive retrieval job identified by JOB_ID in VAULT_NAME
    and save it to FILE_NAME.

    The archive retrieval job must have already been initialized by
    glacier-init-archive-retrieval command.
    """
    return archives.get(vault_name, job_id, file_name)



# A tool to upload and manage archives in AWS Glacier Vaults.
# Copyright (C) 2023 Trapsilo P. Bumi tbumi@thpd.io
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib


def calculate_tree_hash(part, part_size):
    checksums = []
    upper_bound = min(len(part), part_size)
    step = 1024 * 1024  # 1 MB
    for chunk_pos in range(0, upper_bound, step):
        chunk = part[chunk_pos : chunk_pos + step]
        checksums.append(hashlib.sha256(chunk).hexdigest())
        del chunk
    return calculate_total_tree_hash(checksums)


def calculate_total_tree_hash(list_of_checksums):
    tree = list_of_checksums[:]
    while len(tree) > 1:
        parent = []
        for i in range(0, len(tree), 2):
            if i < len(tree) - 1:
                part1 = bytes.fromhex(tree[i])
                part2 = bytes.fromhex(tree[i + 1])
                parent.append(hashlib.sha256(part1 + part2).hexdigest())
            else:
                parent.append(tree[i])
        tree = parent
    return tree[0]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['glacier_upload', 'glacier_upload.utils']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.26.47,<2.0.0', 'click>=8.1.3,<9.0.0']

entry_points = \
{'console_scripts': ['glacier = glacier_upload.cli:glacier_cli']}

setup_kwargs = {
    'name': 'glacier-upload',
    'version': '2.0',
    'description': 'A helper tool to upload and manage archives in AWS Glacier Vaults',
    'long_description': '# glacier-upload\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit)](https://github.com/pre-commit/pre-commit)\n[![pypi](https://img.shields.io/pypi/v/glacier_upload)](https://pypi.org/project/glacier_upload/)\n[![License-GPLv3](https://img.shields.io/github/license/tbumi/glacier-upload)](https://github.com/tbumi/glacier-upload/blob/main/LICENSE)\n\nA helper tool to upload and manage archives in\n[Amazon S3 Glacier](https://docs.aws.amazon.com/amazonglacier/latest/dev/introduction.html)\nVaults. Amazon S3 Glacier is a cloud storage service that is optimized for long\nterm storage for a relatively cheap price.\n\n## Installation\n\nMinimum required Python version is 3.9. To install, run this in your terminal:\n\n```\n$ pip install glacier_upload\n```\n\n## Usage\n\n### Prerequisites\n\nTo upload an archive to Amazon S3 Glacier vault, ensure you have:\n\n-   Created an AWS account\n-   Created an Amazon S3 Glacier vault from the AWS CLI tool or the Management\n    Console\n\n### Uploading an archive\n\nAn upload can be performed by running `glacier upload` followed by the vault\nname and the file name(s) that you want to upload.\n\n```\nglacier upload VAULT_NAME FILE_NAME [FILE_NAME ...]\n```\n\n`FILE_NAME` can be one or more files or directories.\n\nThe script will:\n\n1. Read the file(s)\n2. Consolidate them into a `.tar.xz` archive if multiple `FILE_NAME`s are\n   specified or `FILE_NAME` is one or more directories\n3. Upload the file in one go if the file is less than 100 MB in size, or\n4. Split the file into chunks\n5. Spawn a number of threads that will upload the chunks in parallel. Note that\n   it will not read the entire file into memory, but only parts of the file as\n   it processes the chunks.\n6. Return the archive ID when complete. Consider saving this archive ID in a\n   safe place for retrieval purposes, because Amazon Glacier does not provide a\n   list of archives in realtime. See the "Requesting an inventory" section below\n   for details.\n\nThere are additional options to customize your upload, such as adding a\ndescription to the archive or configuring the number of threads or the size of\nparts. Run `glacier upload --help` for more information.\n\nIf a multipart upload is interrupted in the middle (because of an exception,\ninterrupted manually, or other reason), the script will show you the upload ID.\nThat upload ID can be used to resume the upload, using the same command but\nadding the `--upload-id` option, like so:\n\n```\nglacier upload --upload-id UPLOAD_ID VAULT_NAME FILE_NAME [FILE_NAME ...]\n```\n\n### Retrieving an archive\n\nRetrieving an archive in glacier requires two steps. First, initiate a\n"retrieval job" using:\n\n```\nglacier archive init-retrieval VAULT_NAME ARCHIVE_ID\n```\n\nTo see a list of archive IDs in a vault, see "Requesting an inventory" below.\n\nThen, the retrieval job will take some time to complete. Run the next step to\nboth check whether the job is complete and retrieve the archive if it has been\ncompleted.\n\n```\nglacier archive get VAULT_NAME JOB_ID FILE_NAME\n```\n\n### Requesting an inventory\n\nVaults do not provide realtime access to a list of their contents. To know what\na vault contains, you need to request an inventory of the archive, in a similar\nmanner to retrieving an archive. To initiate an inventory, run:\n\n```\nglacier inventory init-retrieval VAULT_NAME\n```\n\nThen, the inventory job will take some time to complete. Run the next step to\nboth check whether the job is complete and retrieve the inventory if it has been\ncompleted.\n\n```\nglacier inventory get VAULT_NAME JOB_ID\n```\n\n### Deleting an archive, deleting an upload job, creating/deleting a vault, etc.\n\nAll jobs other than uploading an archive and requesting/downloading an inventory\nor archive can be done using the AWS CLI. Those functionalities are not\nimplemented here to avoid duplication of work, and minimize maintenance efforts\nof this package.\n\n## Contributing\n\nContributions and/or bug fixes are welcome! Just make sure you\'ve read the below\nrequirements, then feel free to fork, make a topic branch, make your changes,\nand submit a PR.\n\n### Development Requirements\n\nBefore committing to this repo, install [poetry](https://python-poetry.org/) on\nyour local machine, then run these commands to setup your environment:\n\n```sh\npoetry install\npre-commit install\n```\n\nAll code is formatted with [black](https://github.com/psf/black). Consider\ninstalling an integration for it in your favourite text editor.\n',
    'author': 'Trapsilo Bumi',
    'author_email': 'tbumi@thpd.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tbumi/glacier-upload',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# A tool to upload and manage archives in AWS Glacier Vaults.
# Copyright (C) 2023 Trapsilo P. Bumi tbumi@thpd.io
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import boto3
import click


def init_retrieval(vault_name, format, description):
    glacier = boto3.client("glacier")

    job_params = {"Type": "inventory-retrieval", "Format": format}
    if description is not None:
        job_params["Description"] = description

    click.echo("Sending inventory-retrieval initiation request...")
    try:
        response = glacier.initiate_job(vaultName=vault_name, jobParameters=job_params)
    except glacier.exceptions.ResourceNotFoundException as e:
        raise click.ClickException(e.response["Error"]["Message"])

    click.echo(f"Job initiation request received. Job ID: {response['jobId']}")


def get(vault_name, job_id):
    glacier = boto3.client("glacier")

    click.echo("Checking inventory retrieval status...")
    try:
        response = glacier.describe_job(vaultName=vault_name, jobId=job_id)
    except glacier.exceptions.ResourceNotFoundException as e:
        raise click.ClickException(e.response["Error"]["Message"])

    if response["Action"] != "InventoryRetrieval":
        raise click.ClickException(
            "Job is not an inventory retrieval. Check the Job ID again."
        )

    click.echo(f"Inventory status: {response['StatusCode']}")

    if not response["Completed"]:
        click.echo("Inventory is not completed.")
        return

    click.echo("Retrieving job data...")
    response = glacier.get_job_output(vaultName=vault_name, jobId=job_id)

    if response["contentType"] == "application/json":
        inventory_json = json.load(response["body"])
        click.echo(json.dumps(inventory_json, indent=2))
    else:
        click.echo(response["body"].read())



# A tool to upload and manage archives in AWS Glacier Vaults.
# Copyright (C) 2023 Trapsilo P. Bumi tbumi@thpd.io
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import boto3
import click


def init_retrieval(vault_name, archive_id, description):
    glacier = boto3.client("glacier")

    job_params = {"Type": "archive-retrieval", "ArchiveId": archive_id}
    if description is not None:
        job_params["Description"] = description

    click.echo("Sending archive-retrieval initiation request...")
    try:
        response = glacier.initiate_job(vaultName=vault_name, jobParameters=job_params)
    except glacier.exceptions.ResourceNotFoundException as e:
        raise click.ClickException(e.response["Error"]["Message"])

    click.echo(f"Job initiation request received. Job ID: {response['jobId']}")


def get(vault_name, job_id, file_name):
    glacier = boto3.client("glacier")

    click.echo(f"Checking status of job {job_id} in {vault_name}...")
    try:
        response = glacier.describe_job(vaultName=vault_name, jobId=job_id)
    except glacier.exceptions.ResourceNotFoundException as e:
        raise click.ClickException(e.response["Error"]["Message"])

    if response["Action"] != "ArchiveRetrieval":
        raise click.ClickException(
            "Job is not an archive retrieval. Check the Job ID again."
        )

    click.echo(f"Job status: {response['StatusCode']}")

    if not response["Completed"]:
        click.echo("Job is not completed.")
        return

    if os.path.lexists(file_name):
        click.confirm(
            f"Are you sure you want to overwrite the file {file_name}?", abort=True
        )

    click.echo("Retrieving job data...")
    response = glacier.get_job_output(vaultName=vault_name, jobId=job_id)

    content_length = int(response["ResponseMetadata"]["HTTPHeaders"]["content-length"])
    response_stream = response["body"]
    try:
        download_archive(content_length, response_stream, file_name)
    finally:
        response_stream.close()


def download_archive(content_length, response_stream, file_name):
    click.echo(f"Downloading archive to file {file_name}")
    with open(file_name, "wb") as file:
        if content_length < 4096:
            # Content length is < 4 KB, downloading it in one go
            file.write(response_stream.read())
        else:
            # Download data in chunks of 4 KB
            chunk_size = 4096
            with click.progressbar(
                length=content_length, label="Downloading file"
            ) as bar:
                for chunk in response_stream.iter_chunks(chunk_size):
                    # iter_chunks returns bytes in chunk format
                    # by calling read internally for chunk_size
                    # https://github.com/boto/botocore/blob/51bcacab620bbb35c84157d61b9fed93f2a467f6/botocore/response.py#L125
                    file.write(chunk)
                    bar.update(len(chunk))
